/***********************************************************/
/***********************************************************/
/*                                                                */
/*    Programari MPI                                              */
/*    Nom del programari:   Wave Equation                         */
/*    Descripció:   Equació propagació ones                       */
/*                  https://en.wikipedia.org/wiki/Wave_equation   */
/*                                                                */
/*    Estudiant:   José Antonio Martín Pérez                      */
/*    email:   jmartinperez1@uoc.edu  /  jamartin@protonmail.ch   */
/*                                                                */
/***********************************************************/
/***********************************************************/


# include <stdio.h>
# include <stdlib.h>
# include <math.h>
# include <time.h>

# include "mpi.h"

int main ( int argc, char *argv[] );
double *update ( int id, int p, int n_global, int n_local, int nsteps, 
  double dt );
void collect ( int id, int p, int n_global, int n_local, int nsteps, 
  double dt, double u_local[] );
double dudt ( double x, double t );
double exact ( double x, double t );
void timestamp ( );


int main ( int argc, char *argv[] )
{
  double dt = 0.00125;
  int i_global_hi;
  int i_global_lo;
  int id;
  int n_global = 401;
  int n_local;
  int nsteps = 4000;
  int p;
  double *u1_local;
  double wtime;
/* 
  Inicialització procés MPI
*/
  MPI_Init ( &argc, &argv );
  MPI_Comm_rank ( MPI_COMM_WORLD, &id );
  MPI_Comm_size ( MPI_COMM_WORLD, &p );

  if ( id == 0 ) 
  {
    timestamp ( );
  }

  wtime = MPI_Wtime ( );

/*
  Determinar N_LOCAL
*/
  i_global_lo = (   id       * ( n_global - 1 ) ) / p;
  i_global_hi = ( ( id + 1 ) * ( n_global - 1 ) ) / p;
  if ( 0 < id )
  {
    i_global_lo = i_global_lo - 1;
  }
  n_local = i_global_hi + 1 - i_global_lo;

/* 
  Actualitzar valors N_LOCAL
*/
  u1_local = update ( id, p, n_global, n_local, nsteps, dt );

/* 
  Recol·lecció valors en un array
*/
  collect ( id, p, n_global, n_local, nsteps, dt, u1_local );

/*
  Temps total necessari per fer el càlcul
*/
  wtime = MPI_Wtime ( ) - wtime;
  if ( id == 0 )
  {
    printf ( "\n" );
    printf ( "  Elapsed wallclock time was %g seconds\n", wtime );

  }

/*
  Fi del procés MPI
*/
  MPI_Finalize ( );

/*
  Alliberem memòria
*/
  free ( u1_local );

/*
  Fi
*/
  if ( id == 0 )
  {
    timestamp ( );
  }

  return 0;
}


// Funció UPDATE
// S'avança la solució un determinat nombre de passos.

double *update ( int id, int p, int n_global, int n_local, int nsteps, 
  double dt ) 
{
  double alpha;
  double c;
  double dx;
  int i;
  int i_global;
  int i_global_hi;
  int i_global_lo;
  int i_local;
  int i_local_hi;
  int i_local_lo;
  int j;
  int ltor = 20;
  int rtol = 10;
  MPI_Status status;
  double t;
  double *u0_local;
  double *u1_local;
  double *u2_local;
  double x;
/*
  Es determina el valor ALPHA.
*/
  c = 1.0;
  dx = 1.0 / ( double ) ( n_global - 1 );
  alpha = c * dt / dx;

  if ( 1.0 <= fabs ( alpha ) )
  {
    if ( id == 0 )
    {
      fprintf ( stderr, "\n" );
      fprintf ( stderr, "UPDATE - Warning!\n" );
      fprintf ( stderr, "  1 <= |ALPHA| = | C * dT / dX |.\n" );
      fprintf ( stderr, "  C = %g\n", c );
      fprintf ( stderr, "  dT = %g\n", dt );
      fprintf ( stderr, "  dX = %g\n", dx );
      fprintf ( stderr, "  ALPHA = %g\n", alpha );
      fprintf ( stderr, "  Computation will not be stable!\n" );
    }
    MPI_Finalize ( );
    exit ( 1 );
  }
/*
  El vector de punts N_GLOBAL s'ha de dividir entre els diferents nodes.
*/
  i_global_lo = (   id       * ( n_global - 1 ) ) / p;
  i_global_hi = ( ( id + 1 ) * ( n_global - 1 ) ) / p;
  if ( 0 < id )
  {
    i_global_lo = i_global_lo - 1;
  }

  i_local_lo = 0;
  i_local_hi = i_global_hi - i_global_lo;

  u0_local = ( double * ) malloc ( n_local * sizeof ( double ) );
  u1_local = ( double * ) malloc ( n_local * sizeof ( double ) );
  u2_local = ( double * ) malloc ( n_local * sizeof ( double ) );

  t = 0.0;
  for ( i_global = i_global_lo; i_global <= i_global_hi; i_global++ ) 
  {
    x = ( double ) ( i_global ) / ( double ) ( n_global - 1 );
    i_local = i_global - i_global_lo;
    u1_local[i_local] = exact ( x, t );
  }

  for ( i_local = i_local_lo; i_local <= i_local_hi; i_local++ )
  {
    u0_local[i_local] = u1_local[i_local];
  }
/* 
  Agafa nombre de passos (NSTEPS).
*/
  for ( i = 1; i <= nsteps; i++ )
  {
    t = dt * ( double ) i;
/* 
  Primer pas: Usem la informació inicial de la derivada.
*/
    if ( i == 1 )
    {
      for ( i_local = i_local_lo + 1; i_local < i_local_hi; i_local++ ) 
      {
        i_global = i_global_lo + i_local;
        x = ( double ) ( i_global ) / ( double ) ( n_global - 1 );
        u2_local[i_local] = 
          +         0.5 * alpha * alpha   * u1_local[i_local-1]
          + ( 1.0 -       alpha * alpha ) * u1_local[i_local] 
          +         0.5 * alpha * alpha   * u1_local[i_local+1]
          +                            dt * dudt ( x, t );
      }
    }
/* 
  Després del primer pas, es podem fer servir les estimacions.
*/
    else
    {
      for ( i_local = i_local_lo + 1; i_local < i_local_hi; i_local++ ) 
      {
        u2_local[i_local] = 
          +               alpha * alpha   * u1_local[i_local-1]
          + 2.0 * ( 1.0 - alpha * alpha ) * u1_local[i_local] 
          +               alpha * alpha   * u1_local[i_local+1]
          -                                 u0_local[i_local];
      }
    }
/* 
  Intercanvi d'informació amb el node veí de la part esquerra. 
*/
    if ( 0 < id ) 
    {
      MPI_Send ( &u2_local[i_local_lo+1], 1, MPI_DOUBLE, id - 1, rtol, 
        MPI_COMM_WORLD );
      MPI_Recv ( &u2_local[i_local_lo],   1, MPI_DOUBLE, id - 1, ltor, 
        MPI_COMM_WORLD, &status );
    }
    else
    {
      x = 0.0;
      u2_local[i_local_lo] = exact ( x, t );
    }
/* 
  Intercanvi d'informació amb el node veí de la part dreta.
*/
    if ( id < p - 1 ) 
    {
      MPI_Send ( &u2_local[i_local_hi-1], 1, MPI_DOUBLE, id + 1, ltor, 
        MPI_COMM_WORLD );
      MPI_Recv ( &u2_local[i_local_hi],   1, MPI_DOUBLE, id + 1, rtol, 
        MPI_COMM_WORLD, &status );
    }
    else
    {
      x = 1.0;
      u2_local[i_local_hi] = exact ( x, t );
    }
/*
  Ens movem al següent pas.
*/
    for ( i_local = i_local_lo; i_local <= i_local_hi; i_local++ )
    {
      u0_local[i_local] = u1_local[i_local];
      u1_local[i_local] = u2_local[i_local];
    }
  }
/*
  Alliberem memòria.
*/
  free ( u0_local );
  free ( u2_local );

  return u1_local;
}


// Funció COLLECT
// Recol·lecció de dades i enviament al node mestre

void collect ( int id, int p, int n_global, int n_local, int nsteps, 
  double dt, double u_local[] ) 
{
  int buffer[2];
  int collect1 = 10;
  int collect2 = 20;
  int i;
  int i_global;
  int i_global_hi;
  int i_global_lo;
  int i_local;
  int i_local_hi;
  int i_local_lo;
  int j;
  int n_local2;
  MPI_Status status;
  double t;
  double *u_global;
  double x;

  i_global_lo = (   id       * ( n_global - 1 ) ) / p;
  i_global_hi = ( ( id + 1 ) * ( n_global - 1 ) ) / p;
  if ( 0 < id )
  {
    i_global_lo = i_global_lo - 1;
  }

  i_local_lo = 0;
  i_local_hi = i_global_hi - i_global_lo;
/* 
  Node mestre recull resultats al vector U_GLOBAL.
*/
  if ( id == 0 )
  {
/*
  Creació del vector global.
*/
    u_global = ( double * ) malloc ( n_global * sizeof ( double ) );
/*
  Es copia el resultat del node mestre al vector global.
*/
    for ( i_local = i_local_lo; i_local <= i_local_hi; i_local++ )
    {
      i_global = i_global_lo + i_local - i_local_lo;
      u_global[i_global] = u_local[i_local];
    }
/*
  Es contacta amb cada node
*/
    for ( i = 1; i < p; i++ ) 
    {
/*
  Missatge "collect1", amb l'índex global i el nombre de valors.
*/
      MPI_Recv ( buffer, 2, MPI_INT, i, collect1, MPI_COMM_WORLD, &status );
      i_global_lo = buffer[0];
      n_local2 = buffer[1];

      if ( i_global_lo < 0 )
      {
        fprintf ( stderr, "  Illegal I_GLOBAL_LO = %d\n", i_global_lo );
        exit ( 1 );
      }
      else if ( n_global <= i_global_lo + n_local2 - 1 )
      {
        fprintf ( stderr, "  Illegal I_GLOBAL_LO + N_LOCAL2 = %d\n", 
          i_global_lo + n_local2 );
        exit ( 1 );
      }
/*
  Missatge "collect2", que conté els valors.
*/
      MPI_Recv ( &u_global[i_global_lo], n_local2, MPI_DOUBLE, i, collect2, 
        MPI_COMM_WORLD, &status );
    }
/*
  Es mostra el resultat.
*/
    t = dt * ( double ) nsteps;
    free ( u_global );
  }

/*
  Resta de nodes envien els resultats al node mestre (node 0).
*/
  else
  {
/*
  Missatge "collect1", amb l'índex global i el nombre de valors.
*/
    buffer[0] = i_global_lo;
    buffer[1] = n_local;
    MPI_Send ( buffer, 2, MPI_INT, 0, collect1, MPI_COMM_WORLD );
/*
  Missatge "collect2", que conté els valors.
*/
    MPI_Send ( u_local, n_local, MPI_DOUBLE, 0, collect2, MPI_COMM_WORLD );
  }

  return;
}


//  Funció EXACT
//  Càlcul de la solució exacta.

double exact ( double x, double t )
{
  const double c = 1.0;
  const double pi = 3.141592653589793;
  double value;

  value = sin ( 2.0 * pi * ( x - c * t ) );

  return value;
}


// Funció DUDT
// Evaluació de la derivada parcial.

double dudt ( double x, double t )
{
  const double c = 1.0;
  const double pi = 3.141592653589793;
  double value;

  value = - 2.0 * pi * c * cos ( 2.0 * pi * ( x - c * t ) );

  return value;
}


// Funció TIMESTAMP
// Marca de temps

void timestamp ( )
{
# define TIME_SIZE 40

  static char time_buffer[TIME_SIZE];
  const struct tm *tm;
  time_t now;

  now = time ( NULL );
  tm = localtime ( &now );

  strftime ( time_buffer, TIME_SIZE, "%d %B %Y %I:%M:%S %p", tm );

  printf ( "%s\n", time_buffer );

  return;
# undef TIME_SIZE
}
